// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright (C) 2023 Atmark Techno Inc.
 */

#include <stdbool.h>
#include <stdio.h>
#include <env.h>
#include <i2c.h>
#include <malloc.h>
#include <dm/uclass.h>
#include "board_addition.h"

#define I2C_BUS	(3)

/* EEPROM I2C slave address */
#define EEPROM_ADDR1	(0x57) /* 1st */
#define EEPROM_ADDR2	(0x56) /* 2nd */
#define EEPROM_ADDR3	(0x55) /* 3rd */
#define EEPROM_ADDR4	(0x53) /* 4th */

#define A6E_ADDITIONAL_BOARD_MAX	(4)

/* EEPROM Format */
#define POS_VENDOR	(0x00)
#define POS_PRODUCT	(0x02)

/* Vendor ID's */
#define VENDOR_ATMARK_TECHNO	(0x0001)

/* Product ID's */
#define PRODUCT_ATMARK_TECHNO_DI8AI4	(0x000E)
#define PRODUCT_ERROR			(0xFFFF)

static const unsigned char add_board_i2c_addrs[A6E_ADDITIONAL_BOARD_MAX] = {
	EEPROM_ADDR1,
	EEPROM_ADDR2,
	EEPROM_ADDR3,
	EEPROM_ADDR4,
};

static const char *a6e_di8ai4[A6E_ADDITIONAL_BOARD_MAX] = {
	"armadillo-iotg-a6e-di8ai4-1st.dtbo",
	"armadillo-iotg-a6e-di8ai4-2nd.dtbo",
	"armadillo-iotg-a6e-di8ai4-3rd.dtbo",
	"armadillo-iotg-a6e-di8ai4-4th.dtbo",
};

/* armadillo-iotg-a6e-di8ai4-XXX.dtbo is 34 char's */
#define MAX_DTBO_LEN	(35)

static int eeprom_read(unsigned char i2c_addr, unsigned char addr, void *data,
		       unsigned short size, struct udevice *i2c_dev)
{
	int ret;
	unsigned char *pdata = (unsigned char *)data;

#ifdef CONFIG_DM_I2C
	ret = dm_i2c_read(i2c_dev, addr, pdata, size);
#else
	ret = i2c_read(i2c_addr, addr, pdata, size);
#endif
	if (ret) {
		printf("%s: i2c read failed ret:%d.\n", __func__, ret);
		return ret;
	}
	return 0;
}

static bool is_vendor_atmark_techno(unsigned char i2c_addr, struct udevice *i2c_dev)
{
	unsigned short	vendor_id_eeprom;
	unsigned short	vendor_id;

	if (eeprom_read(i2c_addr, POS_VENDOR, &vendor_id_eeprom, 2, i2c_dev))
		return false;

	vendor_id = be16_to_cpu(vendor_id_eeprom);
	if (vendor_id == VENDOR_ATMARK_TECHNO)
		return true;

	return false;
}

static unsigned short get_product_id(unsigned short i2c_addr,
				     struct udevice *i2c_dev)
{
	unsigned short	product_id_eeprom;
	unsigned short	product_id;

	if (eeprom_read(i2c_addr, POS_PRODUCT, &product_id_eeprom, 2, i2c_dev))
		return PRODUCT_ERROR;

	product_id = be16_to_cpu(product_id_eeprom);
	return product_id;
}

static int remove_additional_dtbo(char *buf)
{
	int i;
	char *p;
	char *p_space;
	char *p_src;
	char *tail;
	int total_len;
	int current_len;

	if (!strlen(buf))
		return 0;

	total_len = strlen(buf) + 1;

	tail = buf + total_len - 1;

	for (i = 0; i < A6E_ADDITIONAL_BOARD_MAX; i++) {
		p = strstr(buf, a6e_di8ai4[i]);
		if (p) {
			p_space = strstr(p, " ");
			if (p_space)
				current_len = p_space - p;
			else
				current_len = strlen(p);

			if (p == buf) {
				/* top */
				if (total_len == (current_len + 1)) {
					/* only this dtbo */
					return 0;
				}
				p_src = buf + current_len + 1;
				memmove(p, p_src, total_len - 1 - (current_len + 1));
				total_len -= current_len + 1;
				tail -= current_len + 1;
				*tail = '\0';
			} else if (total_len == ((p - buf) + current_len + 1)) {
				/* last */
				p--;	/* remove space before */
				total_len -= current_len + 1;
				tail -= current_len + 1;
				*tail = '\0';
			} else {
				/* intermediate */
				p_src = p + current_len + 1;
				memmove(p, p_src, total_len - 1 - (p_src - buf));
				total_len -= current_len + 1;
				tail -= current_len + 1;
				*tail = '\0';
			}
		}
	}
	return total_len;
}

static void check_apply_additional_board(struct udevice *i2c_bus)
{
	char *buf = NULL;
	unsigned short product_id;
	unsigned char i2c_addr;
	struct udevice *i2c_dev = NULL;
	int i;
	int len = 0;
	int ret;
	bool is_continue = true;
	bool found = false;

	buf = env_get("fdt_overlays");
	if (buf) {
		buf = strdup(buf);
		len = remove_additional_dtbo(buf);
		if (!len) {
			free(buf);
			buf = NULL;
		}
	}

	for (i = 0; i < A6E_ADDITIONAL_BOARD_MAX; i++) {
		i2c_addr = add_board_i2c_addrs[i];

#if CONFIG_DM_I2C
		ret = dm_i2c_probe(i2c_bus, i2c_addr, 0, &i2c_dev);
#else
		ret = i2c_probe(i2c_addr);
#endif
		if (ret)
			break;

		if (!is_vendor_atmark_techno(i2c_addr, i2c_dev))
			break;

		product_id = get_product_id(i2c_addr, i2c_dev);
		switch (product_id) {
		case PRODUCT_ATMARK_TECHNO_DI8AI4:
			if (len) {
				/* put space for after 2nd dtbo */
				buf = realloc(buf, len + 1);
				snprintf(&buf[len - 1], 2, " ");
				len++;
			}

			if (buf) {
				buf = realloc(buf,
					      len + strlen(a6e_di8ai4[i]) + 1);
				len += snprintf(&buf[len - 1],
						strlen(a6e_di8ai4[i]) + 1,
						"%s", a6e_di8ai4[i]);
			} else {
				buf = strdup(a6e_di8ai4[i]);
				len += strlen(a6e_di8ai4[i]) + 1;
			}
			found = true;
			break;
		default:
			/* The +Di8+Ai4 board is not recognized unless it is
			 * added in order from the 1st.
			 */
			is_continue = false;
			break;
		}

		if (!is_continue)
			break;
	}

	if (len && found)
		env_set("fdt_overlays", buf);

	free(buf);
}

static bool setup_i2c(struct udevice **i2c_bus)
{
	int ret;

#ifdef CONFIG_DM_I2C
	ret = uclass_get_device_by_seq(UCLASS_I2C, I2C_BUS, i2c_bus);
#else
	ret = i2c_set_bus_num(I2C_BUS);
#endif
	if (ret)
		return false;

	return true;
}

static int check_fdt_autodetect(void)
{
	char *autodetect_str = env_get("fdtautodetect");

	if (autodetect_str &&
	    (strcmp(autodetect_str, "yes") == 0)) {
		return 1;
	}

	return 0;
}

int detect_additional_boards(void)
{
	struct udevice *i2c_bus = NULL;

	if (!check_fdt_autodetect())
		return 0;

	if (!setup_i2c(&i2c_bus))
		return 0;

	check_apply_additional_board(i2c_bus);

	return 0;
}
